import asyncio
import os
import shutil
from pathlib import Path
import requests
import subprocess
from py7zr import unpack_7zarchive
import json
import orjson
import pickle
import numpy as np
import pandas as pd
import aiofiles
import io
from dl2050utils.core import listify

Path.ls = lambda x: list(x.iterdir())

# ################################################################################
# cp, rm
# ################################################################################

def cp(p1, p2):
    """ Copies recursively p1 to p2 with overwrite. If p1 is a directory, copies recursively. """
    p1,p2 = Path(p1),Path(p2)
    if not p1.exists():
        print(f'File {p1} not found')
        return 1
    try:
        if p1.is_file():
            shutil.copyfile(p1, p2)
            return 0
        if p2.is_dir():
            shutil.rmtree(p2)
        shutil.copytree(p1, p2)
    except Exception:
        return 1
    
def rm(p):
    " Removes file or directory p. If p is a directory, removes recursively. "
    try: 
        p = Path(p)
        if p.is_file():
            p.unlink()
            return 0
        if p.is_dir():
            return shutil.rmtree(p)
    except Exception:
        return 1

# ################################################################################
# sh_run, run_asyc
# ################################################################################

def sh_run(cmd):
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    (stdout, stderr) = proc.communicate()
    stdout, stderr = stdout.decode("utf-8"), stderr.decode("utf-8")
    if proc.returncode!=0: raise RuntimeError(f'sh_run command \n{cmd}\n exit with error code {proc.returncode}:\n{stderr}')
    return (proc.returncode, stdout)

async def run_asyc(*args):
    try:
        process = await asyncio.create_subprocess_exec(*args, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
        stdout, stderr = await process.communicate()
        return process.returncode, stdout.decode(), stderr.decode()
    except:
        return -1, '', 'Command not found'
    
# ################################################################################
# get_all_files, get_dir_files, get_dir_dirs
# ################################################################################

def get_all_files(p, filetypes=None):
    filetypes = [e.lower() for e in listify(filetypes)]
    filetypes = [f'.{e}' if e[0]!='.' else e for e in filetypes]
    files = [p1 for p1 in p.rglob('*')]
    if len(filetypes): files = [f for f in files if f.suffix.lower() in filetypes]
    return files

# Draft
def get_dir_files(path, types=None):
    return [f for f in path.glob('**/*') if f.is_file() and (types==None or f.suffix[1:] in types)]

# Draft
def get_dir_dirs(path):
    return [d for d in path.glob('**/*') if d.is_dir()]
    
# ################################################################################
# zip_get_registered_formats, zip_register_formats, iszip, dozip, unzip
# ################################################################################

def zip_get_registered_formats():
    return sum([e[1] for e in shutil.get_unpack_formats()], [])
    
def zip_register_formats():
    zips = zip_get_registered_formats()
    if '.7z' not in zips:
        print('Registering .7z')
        shutil.register_unpack_format('7zip', ['.7z'], unpack_7zarchive)
    if '.hsz' not in zips:
        print('Registering .hsz')
        shutil.register_unpack_format('7zip', ['.hsz'], unpack_7zarchive)

def iszip(p):
    return (Path(p)).suffix in zip_get_registered_formats()

def dozip(p, remove=False):
    p = Path(p)
    if not p.exists():
        print(f'dozip: file not found: {p}')
        return 1
    try: 
        shutil.make_archive(p, 'zip', p)
        if remove:
            if rm(p): return 1
        return 0
    except Exception:
        return 1
    
def unzip(p, remove=False):
    p = Path(p)
    if not p.is_file():
        print(f'unzip: file not found: {p}')
        return 1
    p_dir = p.with_suffix('')
    try: 
        shutil.unpack_archive(p, p_dir)
        if remove:
            if rm(p): return 1
        return 0
    except Exception as exc:
        print(f'Exception: {exc}')
        return 1
    
def unzip_tree(p):
    p_zips = [p1 for p1 in get_all_files(p) if iszip(p1)]
    n = 0
    for p1 in p_zips:
        print(f'unzip_tree: unzipping {p1}')
        if unzip(p1, remove=True):
            print(f'unzip_tree: error unzipping {p1}')
            continue
        n += 1
    print(f'unzip_tree: unzipped {n}/{len(p_zips)} files')

# ################################################################################
# pickle_save (pickle_dump), pickle_load
# np_save, np_load, np_load_async
# df_save, df_load
# json_load (json_parse), json_dump, read_json
# json_saves, json_loads
# ################################################################################

def pickle_save(p, d):
    p = Path(p)
    p = p.with_suffix('.pickle')
    try:
        with open(p, 'wb') as f:
            pickle.dump(d, f)
        return 0
    except Exception:
        return 1

# Deprecated
def pickle_dump(p, d): return pickle_save(p, d)
    
def pickle_load(p):
    p = Path(p)
    p = p.with_suffix('.pickle')
    if not p.is_file():
        return None
    try:
        with open(p, 'rb') as f:
            d = pickle.load(f)
        return d
    except Exception:
        return None
    
def np_save(p, d, allow_pickle=True):
    p = Path(p)
    p = p.with_suffix('.npy')
    try:
        np.save(p, d, allow_pickle=allow_pickle)
        return 0
    except Exception:
        return 1
    
def np_load(p, allow_pickle=True):
    p = Path(p)
    p = p.with_suffix('.npy')
    if not p.is_file():
        return None
    try:
        if allow_pickle:
            d = np.load(p, allow_pickle=allow_pickle)
            if d is None: return None
            return d.item()
        else:
            return np.load(p)
    except Exception:
        return None
    
async def np_load_async(p, allow_pickle=False):
    p = Path(p)
    p = p.with_suffix('.npy')
    if not p.is_file():
        return None
    try:
        async with aiofiles.open(p, mode='rb') as f:
            buffer = await f.read()
        f = io.BytesIO(buffer)
        if allow_pickle:
            d = np.load(f, allow_pickle=allow_pickle)
            if d is None: return None
            return d.item()
        else:
            return np.load(f, allow_pickle=allow_pickle)
    except Exception:
        return None
    
def df_save(p, df):
    p = Path(p)
    p = p.with_suffix('.feather')
    try:
        df.to_feather(p)
        return 0
    except Exception:
        return 1
    
def df_load(p):
    p = Path(p)
    p = p.with_suffix('.feather')
    if not p.is_file():
        return None
    try:
        df = pd.read_feather(p)
        return df
    except Exception:
        return None

def json_save(p, d):
    p = Path(p)
    p = p.with_suffix('.json')
    try:
        with open(p, 'wb') as f:
            f.write(orjson.dumps(d))
        return 0
    except Exception:
        return 1
    
def json_load(p):
    p = Path(p)
    p = p.with_suffix('.json')
    if not p.is_file():
        return None
    try:
        with open(p, 'rb') as f:
            return orjson.loads(f.read())
    except Exception:
        return None

def json_dumps(o):
    try:
        return orjson.dumps(o, option=orjson.OPT_SERIALIZE_NUMPY).decode()
    except:
        return None

def json_loads(s):
    if s is None or s=='': return None
    try:
        o = orjson.loads(s)
        if o=={}: return None
        return o
    except:
        return None

# To be reviwed (used in PW) and renamed to json_load
def read_json(p): return json_load(p)
    
# ################################################################################
# download_url, download
# ################################################################################

def download_url(url, dest, overwrite=False, pbar=None, show_progress=True, chunk_size=1024*1024, timeout=4, retries=5):
    "Download `url` to `dest` unless it exists and not `overwrite`."
    # wget -c <url> && tar -zxvf <fname>
    if os.path.exists(dest) and not overwrite: return
    s = requests.Session()
    s.mount('http://',requests.adapters.HTTPAdapter(max_retries=retries))
    u = s.get(url, stream=True, timeout=timeout)
    try: file_size = int(u.headers["Content-Length"])
    except: show_progress = False

    with open(dest, 'wb') as f:
        nbytes = 0
        # if show_progress: pbar = progress_bar(range(file_size), auto_update=False, leave=False, parent=pbar)
        try:
            for chunk in u.iter_content(chunk_size=chunk_size):
                nbytes += len(chunk)
                if show_progress: pbar.update(nbytes)
                f.write(chunk)
        except requests.exceptions.ConnectionError as e:
            print(f'\n Download of {url} has failed after {retries} retries.\nError: {e}')
            return False
    return True

def download(url, fname=None, ext='.tgz'):
    "Download `url` to destination `fname`."
    if fname is None:
        fname = Path(f"{url.split('/')[-1]}{ext}")
    os.makedirs(fname.parent, exist_ok=True)
    if not fname.exists():
        print(f'Downloading {url}')
        if not download_url(f'{url}{ext}', fname):
            return None
    return fname



"""
restapp.py
App Class

Defines the logic for the REST App, including startup, shudown, Auth routes, downstream services routes and Custom Application routes.

The Auth routes are defined in the Auth module.

The downstream services routes include:
    /api/contact:
    /api/get_meta:
    /api/download_general
    /api/download_user/{fname}
    /api/download_req'
    /api/upload_req
    /api/publish_job
    /api/get_jobs
    /api/get_pending_jobs
    /api/get_job

The Custom Application are defined in the routes dict with the following structure:
    routes = {
        <route_path_string>: <route_defenition>,
        ...
    }
    <route_defenition> is a dict with keys 'f','args' and 'kwargs', with the function name implementing the route along with
    corresponding args and kwargs as lists of strings.

    For example:
        routes = {
            '/api/health': {'f': health, 'args':[], 'kwargs':[]},
            ...
        }

Every custom function implementing the route must observe the following:
    1) All inputs parsed from the Request as function parametes that must be defined in the lists args and kwargs above;
    2) All returns should be a dict with the format {status:<status>, ...}:
        - <status> should be either 'OK' or 'ERROR'
        - If 'status':'OK' extra attributes may be present and will be sent to the client
        - If 'status':'ERROR', attribute 'error_msg' should be present, optionally along with 'error_code'
        - If status is not present, 'status':'OK' is automatically inserted;
        - If the return is not a dict, a dict is created with {'status':'OK', 'result':<return>}
        - If None is returned, a dict is created with only {'status':'OK'}
        Examples:
            return {'status':'ERROR', 'error_msg':'Error message', 'error_code':'Error code'} # Client gets what is returned
            return {'data':[arg1, kwarg1]} # Client gets {'status':'OK', 'data':[arg1, kwarg1]}
            return [1,2,3] # Client gets {'status':'OK', 'result':[1, 2, 3]}
            return None # Client gets {'status':'OK'}
            return {'status':'ERROR', 'error_msg':'Error message', 'error_code':'Error code'} # Client gets what is returned
    3) All normal returns are converted into HTTP responses with HTTP status 200 and the return data as JSON

    4) It is possible to raise an exception through log_and_raise_app(msg), that is converted in an error return
       of the form {'status':'ERROR', 'error_type':'APP', 'error_msg':<msg>} and then sent as an HTTP response
       with HTTP status 400 and the error data as JSON

    5) If the function has an uncatched exception (bug), the App catches it and converts in the following response:
       {'status':'ERROR', 'error_type':'EXCEPTION', 'error_msg': <exception_details>} and then sent as an HTTP response
       with HTTP status 400 and the error data as JSON


"""

import time
from datetime import datetime, timezone
from pathlib import Path
from starlette.responses import FileResponse
from starlette.routing import Route
from starlette.authentication import requires
from dl2050utils.core import oget
from dl2050utils.restutils import HTTPException, log_and_raise_exception, log_and_raise_rest, log_and_raise_service, rest_ok,\
                                  enforce_required_args, get_optional_args, get_meta, get_upload_url, get_download_url

class App():

    def __init__(self, cfg, LOG, NOTIFY, db, mq, auth, path, routes=[], appstartup=None, perm=None):
        self.path = path
        self.cfg,self.LOG,self.NOTIFY,self.db,self.mq,self.auth = cfg,LOG,NOTIFY,db,mq,auth
        self.routes,self.appstartup,self.perm = routes,appstartup,perm
        self.d = {'LOG':LOG, 'path':path, 'db':db, 'mq':mq}
        self.fs_secret = oget(self.cfg,['fs','secret'])
        if self.fs_secret is None:
            self.LOG(3, 0, label='APP', label2='REST',  msg='fs_secret not found')

    async def startup(self):
        model,meta = oget(self.cfg,['app','model']),None
        if model is not None: meta = await get_meta(self.path, self.db, model)
        self.d['meta'] = meta
        if self.appstartup is None: return False
        return await self.appstartup(self.d)

    def shutdown(self):
        self.LOG(2, 0, label='APP', label2='shutdown', msg='OK')
        return False   

    def get_routes(self):
        BASE_ROUTES = [
            Route('/api/contact', endpoint=self.contact, methods=['POST']),
            Route('/api/get_meta', endpoint=self.get_meta, methods=['GET']),
            Route('/api/download_general', endpoint=self.download_general, methods=['GET']),
            Route('/api/download_user/{fname}', endpoint=self.download_user, methods=['GET']),
            Route('/api/download_req', endpoint=self.download_req, methods=['POST']),
            Route('/api/upload_req', endpoint=self.upload_req, methods=['POST']),
            Route('/api/publish_job', endpoint=self.publish_job, methods=['POST']),
            Route('/api/get_jobs', endpoint=self.get_jobs, methods=['POST']),
            Route('/api/get_pending_jobs', endpoint=self.get_pending_jobs, methods=['POST']),
            Route('/api/get_job', endpoint=self.get_job, methods=['POST']),
        ]
        APP_ROUTES = [Route(e, endpoint=self.app_route, methods=['POST']) for e in self.routes]
        return BASE_ROUTES + APP_ROUTES

    async def db_increase(self, tbl, k, kv, c, prefix=''):
        row = await self.db.select_one(tbl, {k: kv})
        if row is None or len(row)==0:
            log_and_raise_service(self.LOG, label='AUTH', label2=prefix, msg='DB increase error: select')
        row[c] += row[c]+1
        res = await self.db.update(tbl, row)
        if res:
            log_and_raise_service(self.LOG, label='AUTH', label2=prefix, msg='DB increase error: update')

    async def contact(self, request):
        data = await request.json()
        d = enforce_required_args(self.LOG, data, ['email','name','msg'], label='AUTH', label2='check_user')
        d['ts']: datetime.now(timezone.utc)
        err = await self.db.insert('contacts', d)
        if err:
            log_and_raise_service(self.LOG, label='REST', label2='contact', msg='DB access error')
        html = '<h1>Name</h1><p>{d["email"]</p><h1>Email</h1><p>{d["email"]</p><h1>Message</h1><p>{d["msg"]</p>}'
        self.notify.send_mail_async(d['email'], subject='Contact from Website (cardiolife.global)', html=html)
        return rest_ok()

    @requires('authenticated')
    async def get_meta(self, request):
        u = await self.auth.check_auth(request)
        return rest_ok(self.d['meta'])

    @requires('authenticated')
    async def download_general(self, request):
        u = await self.auth.check_auth(request)
        uid = u['uid']
        fname = request.path_params['fname']
        path = self.d['path']
        p = Path(f'{path}/private/{uid}/{fname}')
        return FileResponse(str(p), media_type='application/vnd.ms-excel', filename=fname)

    @requires('authenticated')
    async def download_user(self, request):
        u = await self.auth.check_auth(request)
        uid = u['uid']
        fname = request.path_params['fname']
        path = self.d['path']
        p = Path(f'{path}/tmp/{int(time.time())//86400}-{uid}-{fname}')
        if not p.is_file():
            log_and_raise_service(self.LOG, label='REST', label2='download_user', msg='File requested for download not found')
        return FileResponse(str(p), media_type='application/vnd.ms-excel', filename=fname)

    @requires('authenticated')
    async def download_req(self, request):
        if not self.fs_secret:
            log_and_raise_service(self.LOG, label='REST', label2='download_req', msg='Upload secret not defined')
        u = await self.auth.check_auth(request)
        # self.db_increase('users', 'uid', u['uid'], 'downloads', 'download_req')
        data = await request.json()
        args = enforce_required_args(self.LOG, data, ['bucket','fname'], label='REST', label2='download_req')
        self.perm(self.d, u, request.url.path, args)
        res = {'url': get_download_url(self.fs_secret, args['bucket'], args['fname'], timeout=7*24*3600)}
        return rest_ok(res)

    @requires('authenticated')
    async def upload_req(self, request):
        if not self.fs_secret:
            log_and_raise_service(self.LOG, label='REST', label2='upload_req', msg='Upload secret not defined')
        u = await self.auth.check_auth(request)
        # self.db_increase('users', 'uid', u['uid'], 'uploads', 'upload_req')
        data = await request.json()
        args = enforce_required_args(self.LOG, data, ['bucket','fname','size'], label='REST', label2='upload_req')
        self.perm(self.d, u, request.url.path, args)
        res = {'url': get_upload_url(self.fs_secret, args['bucket'], args['fname'], args['size'], timeout=7*24*3600)}
        return rest_ok(res)

    @requires('authenticated')
    async def publish_job(self, request):
        u = await self.auth.check_auth(request)
        uid = u['uid']
        # To be replaced -> uid will be the email
        email = await self.auth.get_email_from_uid(uid)
        data = await request.json()
        args = enforce_required_args(self.LOG, data, ['q','payload'], label='REST', label2='publish_job')
        q,payload = args['q'],args['payload']
        self.perm(self.d, u, request.url.path, args)
        jid = await self.mq.publish(q, email, payload)
        if jid is None :
            log_and_raise_service(self.LOG, label='REST', label2='publish_job', msg='MQ publish error')
        self.LOG(2, 0, label='APP', label2='publish_job', )
        return rest_ok({'jid': jid})

    @requires('authenticated')
    async def get_jobs(self, request):
        u = await self.auth.check_auth(request)
        uid = u['uid']
        # To be replaced -> uid will be the email
        email = await self.auth.get_email_from_uid(uid)
        data = await request.json()
        kwargs = get_optional_args(data, ['jstatus'])
        jobs = await self.mq.get_jobs({email:email, **kwargs})
        return rest_ok(jobs)

    @requires('authenticated')
    async def get_pending_jobs(self, request):
        u = await self.auth.check_auth(request)
        uid = u['uid']
        # To be replaced -> uid will be the email
        email = await self.auth.get_email_from_uid(uid)
        jobs = await self.mq.get_pending_jobs(email)
        return rest_ok(jobs)

    @requires('authenticated')
    async def get_job(self, request):
        u = await self.auth.check_auth(request)
        data = await request.json()
        args = enforce_required_args(self.LOG, data, ['jid'], label='REST', label2='get_job')
        job = await self.mq.get_job(args['jid'])
        return rest_ok(job)

    @requires('authenticated')
    async def app_route(self, request):
        if request.url.path not in self.routes:
            log_and_raise_rest(self.LOG, label='REST', label2=f'app_route {request.url.path}', msg='url path not found')
        d = self.routes[request.url.path]
        cb,args,kwargs = d['f'],d['args'],d['kwargs']
        u = await self.auth.check_auth(request)
        data = await request.json()
        args2 = enforce_required_args(self.LOG, data, args, label='REST', label2='app_route')
        kwargs2 = get_optional_args(data, kwargs)
        self.perm(self.d, u, request.url.path, {**args2, **kwargs2})
        try:
            res = await cb(self.d, u, *[args2[e] for e in args2], **kwargs2)
        except HTTPException as exc:
            raise
        except Exception as exc:
            log_and_raise_exception(self.LOG, label='REST', label2='app_route', msg=str(exc))
        return rest_ok(res)



"""
    wsgi.py
    Sets up an Web Server Gateway Interface (WSGI) backend.
    
"""

import traceback
import time
import cgi
import json
from http.server import BaseHTTPRequestHandler, HTTPServer
from dl2050utils.restutils import enforce_required_args, get_optional_args

DEBUG = True

class Handler(BaseHTTPRequestHandler):
    def log_message(self, format, line, return_code, _):
        pass
        # msg = f'{self.command} {return_code}'
        # if self.error: msg+=f' {self.error}'
        # # msg += f' {self.client_address}'
        # t = self.t0 if hasattr(self, 't0') else 0
        # level = 2 if return_code=='200' else 4
        
    def _set_headers(self):
        self.send_response(200)
        self.send_header('Content-type', 'application/json')
        self.end_headers()
        
    def do_HEAD(self):
        self.t0 = time.time()
        self._set_headers()
        
    def do_GET(self):
        self.t0 = time.time()
        self.send_response(400)
        self.end_headers()

    def do_POST(self):
        try:
            self.t0 = time.time()
            ctype, pdict = cgi.parse_header(self.headers.get('content-type'))
            if ctype!='application/json':
                raise Exception('Content-type error')
            if self.path not in self.routes:
                raise Exception('Invalid route')
            length = int(self.headers.get('content-length'))
            payload = json.loads(self.rfile.read(length))
            route = self.routes[self.path]
            cb,args,kwargs = route['f'],route['args'],route['kwargs']
            args2 = enforce_required_args(None, payload, args, label='WSGI', label2='do_POST')
            kwargs2 = get_optional_args(payload, kwargs)
            res = cb(self.d, *[args2[e] for e in args2], **kwargs2)
            if res is None: res = {}
            if type(res)!=dict: res = {'result':res}
            if 'status' not in res: res['status'] = 'OK'
            res = json.dumps(res)
            res = res.encode()
            self.LOG(2, time.time()-self.t0, label='WSGIServer-OK', label2=self.path, msg=f'OK {self.client_address}')
            self._set_headers()
            self.wfile.write(res)
        except Exception as err:
            self.error = str(err)
            trace = traceback.format_exc() if DEBUG else ''
            self.send_response(400)
            self.send_header('Content-type','application/json')
            self.end_headers()
            res = {'status':'ERROR', 'error_type':'EXCEPTION', 'error_msg':str(err), 'trace':trace} 
            self.LOG(4, time.time()-self.t0, label='WSGIServer', label2=self.path, msg=res)
            res = json.dumps(res)
            res = res.encode()
            self.wfile.write(res)


def HandlerFactory(routes, d):
    class ExtendedHandler(Handler):
        def __init__(self, *args, **kwargs):
            super(ExtendedHandler, self).__init__(*args, **kwargs)
    ExtendedHandler.routes = routes
    ExtendedHandler.d = d
    ExtendedHandler.LOG = d['LOG']
    ExtendedHandler.error = None
    return ExtendedHandler

class WSGIServer():
    def __init__(self, routes, startup, d={}, secret=None, host='', port=6000):
        self.routes,self.startup = routes,startup
        self.secret,self.host,self.port = secret, host,port
        self.d = d
        self.LOG = d['LOG']
        
    def run(self, host='', port=6000):
        try:
            if self.startup(self.d):
                return True
        except Exception as err:
            self.LOG(4, 0, label='WSGIServer', label2='STARTUP EXCEPTION', msg=str(err))
            raise RuntimeError('WSGIServer STARTUP EXCEPTION')
        Handler = HandlerFactory(self.routes, self.d)
        httpd = HTTPServer((host, port), Handler)
        self.LOG(2, 0, label='WSGIServer', label2='RUN', msg=f'Listenning on {self.port}')
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            pass
        httpd.server_close()



from pathlib import Path
from dotenv import load_dotenv
import yaml

# ################################################################################
# config_load
# ################################################################################

def find_yml(name):
    ps = [Path(f'./{name}'), Path(f'/run/secrets/{name}')]
    for p in ps:
        if p.exists(): return p
    return None

def config_load(name=None):
    """
        Load config yml file.
        File name is f'config-{name}.yml', or config.yml by default.
        Checks for file first in the current folder then in /run/secrets folder.
    """
    load_dotenv()
    name = 'config.yml' if name is None else f'config-{name}.yml'
    fname = find_yml(name)
    if fname is None:
        raise RuntimeError('Unable to find config file')
    with open(str(fname),'r') as f:
        try:
            cfg = yaml.safe_load(f)
        except yaml.YAMLError as err:
            raise RuntimeError(f'Unable to load config file: {str(err)}')
    return cfg



"""
    rest.py
    Sets up the Asynchronous Server Gateway Interface (ASGI) Server backend

    Start an univorn server based on a set of downstream services: AppLog, Notify, DB, MQ, Auth and App.

    Serves all the routes included in three groups:
        1) Auth routes define in the Auth instance
        2) Base routes defined in the App instance for general services
        3) Custom Application routes passed to the App instance from upper layer

    Errors are handled in layers
        1)  EXCEPTION in ASGI internal:
            Bugs inside backend module, should never arise
            HTTP return code: 500
            TEXT response set to traceback if debug mode is on
        2)  EXCEPTION in Middleware or endpoint:
            Bugs inside Middleware or Application modules, should never arise especially in Middleware
            HTTP return code: 400
            JSON response: {status:'ERROR', error_type:'EXCEPTION', error_msg:<err_msg>}
        3)  REST ERROR in accessing Endpoints:
            Request Errors with bad formed routes or missing or wrong arguments
            HTTP return code: 400
            JSON response: {status:'ERROR', error_type:'REST', error_msg:<err_msg>}
        3)  SERVICE ERROR in downstream services:
            Errors in well formed accesses to downstram services like DB, should never happer
            HTTP return code: 400
            JSON response: {status:'ERROR', error_type:'SERVICE', error_msg:<err_msg>}
        4) APP ERROR in App services:
            Errors in well formed accesses to App services, like missing users. Common and normnal.
            HTTP return code: 400
            JSON response: {status:'ERROR', error_type:'APP', error_msg:<err_msg>}
            Alternatively App services may return a normal response with extra information to deal with the error.
        5) Normal responses:
            HTTP return code: 200
            JSON response: {status:'OK', [msg:...,], ...}
"""

import traceback
import time
import uvicorn
from starlette.applications import Starlette
from starlette.responses import JSONResponse
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette_jwt import JWTAuthenticationBackend
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.responses import JSONResponse
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from dl2050utils.core import oget
from dl2050utils.env import config_load
from dl2050utils.log import AppLog
from dl2050utils.com import Notify, read_address_book
from dl2050utils.db import DB
from dl2050utils.mq import MQ
from dl2050utils.auth import Auth
from dl2050utils.restutils import HTTPException
from dl2050utils.restapp import App

DEBUG = True

class ASGI_Server():
    def __init__(self, name, path, routes, appstartup, perm, qs):
        cfg = config_load(name)
        LOG = AppLog(cfg)
        NOTIFY = Notify(cfg=cfg, address_book=read_address_book())
        dbname = oget(cfg, ['db','dbname'], 'postgres')
        db = DB(cfg=cfg, log=LOG, dbname=dbname)
        mq = MQ(LOG, db, qs, cfg)
        auth = Auth(cfg, LOG, NOTIFY, db)
        app = App(cfg, LOG, NOTIFY, db, mq, auth, path, routes, appstartup, perm)
        self.LOG,self.db,self.mq,self.auth,self.app = LOG,db,mq,auth,app
        self.port = oget(cfg, ['rest','port'], 5000)
        self.exception_handlers = {
            HTTPException: self.http_exception,
            404: self.http_exception,
            Exception: self.server_error_exception,
            500: self.server_error_exception,
        }
        auth_secret = cfg['rest']['auth_secret']
        self.middleware = [
            Middleware(RestStartMiddleware),
            Middleware(CORSMiddleware, allow_origins=['*'], allow_methods=['*'], allow_headers=['*'], expose_headers=['*']),
            Middleware(AuthenticationMiddleware, backend=JWTAuthenticationBackend(secret_key=auth_secret, prefix='Bearer', username_field='uid')),
            Middleware(RestEndMiddleware, LOG=self.LOG, auth=auth),
        ]
        self.routes = auth.get_routes()+app.get_routes()

    async def http_exception(self, request, exc):
        request.state.error = True
        error_type = exc.error_type if hasattr(exc, 'error_type') else 'REST'
        res = {'status':'ERROR', 'error_type':error_type, 'error_msg':exc.detail}
        return JSONResponse(res, status_code=exc.status_code)

    async def server_error_exception(self, request, exc):
        self.LOG(4, 0, label='REST', label2='SERVER_EXCEPTION', msg=exc.detail)
        return JSONResponse({'status': 'SERVER_EXCEPTION', 'error_msg': exc.detail}, status_code=exc.status_code)
        
    async def startup(self):
        if await self.db.startup(): raise('DB startup error')
        if await self.mq.startup(): self.LOG(3, 0, label='REST', label2='STARTUP', msg='MQ not available')
        if await self.auth.startup(): raise('Auth startup error')
        if await self.app.startup(): raise('App startup error')
        self.LOG(2, 0, label='REST', label2='STARTUP', msg=f'OK (port {self.port})')
        return False
            
    async def shutdown(self):
        self.LOG(2, 0, label='REST', label2='SHUTDOWN')
        
    def run(self, port=None):
        rest = Starlette(
            debug=DEBUG,
            exception_handlers=self.exception_handlers,
            middleware=self.middleware,
            on_startup=[self.startup],
            on_shutdown=[self.shutdown],
            routes=self.routes,
        )
        port = port or self.port
        uvicorn.run(rest, port=port, host='0.0.0.0', log_level='critical')

class RestStartMiddleware(BaseHTTPMiddleware):
    def __init__(self, app):
        super().__init__(app)
    async def dispatch(self, request, call_next):
        request.state.error = False
        request.state.t0 = time.time()
        return await call_next(request)

class RestEndMiddleware(BaseHTTPMiddleware):
    def __init__(self, app, LOG=None, auth=None):
        super().__init__(app)
        self.LOG,self.auth=LOG,auth
    async def dispatch(self, request, call_next):
        url = request.url.path
        uid = -1 if request.user.display_name=='' else request.user.display_name
        try:
            response = await call_next(request)
        except Exception as exc:
            error_msg = str(exc)
            log_msg = {'error_msg':error_msg, 'trace':traceback.format_exc()} if DEBUG else error_msg
            t = time.time()-request.state.t0
            self.LOG(4, t, label='REST EXCEPTION', label2=url, msg=log_msg)
            return JSONResponse({'status':'ERROR', 'error_type':'EXCEPTION', 'error_msg':error_msg}, status_code=400)
        if not request.state.error:
            t = time.time()-request.state.t0
            await self.auth.register_requests(url, uid, t)
            self.LOG(2, t, label='REST', label2=url, msg={'userid':uid})
        return response
